///////////////////////////////////////////////////////////
//                                                       //
//          Advanced map editor plugin for VIDE          //
//                     - by DanB -                       //
//               - Released 2008-05-13 -                 //
//                                                       //
///////////////////////////////////////////////////////////
/* Thanks goes to Alberto Covarrubias for VIDE and for   */
/* the original plugin example code                      */

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <commctrl.h>
#include "resource.h"

// Some defines for neater code further down
#define TIMER_ANI 1
// Row and column coordinates for map and char indexes
#define X(index) (index & 0x3F)
#define Y(index) (index / 64)
#define chX(index) (index & 0x1F)
#define chY(index) (index / 32)
// Screen coordinates for map and char indexes
#define mapX(index) (X(index) * 8)
#define mapY(index) (Y(index) * 8)
#define chmapX(index) (chX(index) * 16)
#define chmapY(index) (chY(index) * 16)

// Global variables are always nice...
HINSTANCE hDllModule;
// The four palettes
BYTE color[4][4] = {{0, 85, 170, 255}, {0, 0, 85, 170}, {255, 170, 85, 0}, {0, 0, 255, 170}};
// Our clipboard format and dialog result
int iFormat, dlgResult;
// Index coordinates for the selection boxes
unsigned int selBegin, selEnd, selWidth, selHeight, selA, selB;
unsigned int selBeginCh, selEndCh, selWidthCh, selHeightCh, selACh, selBCh;
TCHAR mapName[60];
BYTE charSet[32768];
WORD bgMap[4096];
HDC charDC[5], mapDC;
HBITMAP charBMP[5], mapBMP;
// The map and charset areas on screen
RECT mapRECT = {64, 0, 576, 512};
RECT chRECT  = {64, 514, 576, 770};
// An animation node
struct aniframe {
	WORD charNo;
	aniframe *next;
};
aniframe aniframes[512];
aniframe *currFrame[512];
// An undo node
struct undo {
	WORD bgMap[4096];
	undo *prev;
};
undo *lastUndo;

BOOL APIENTRY DllMain(HINSTANCE hModule, DWORD ul_reason_for_call, LPVOID lpReserved) {
   hDllModule = hModule;
   
   switch (ul_reason_for_call)
   {
      case DLL_PROCESS_ATTACH:
      case DLL_THREAD_ATTACH:
      case DLL_THREAD_DETACH:
      case DLL_PROCESS_DETACH:
      break;
   }
   return TRUE;
}

// Gets the plugin name for display in VIDE's plugin menu
int GetPluginName(BYTE * name, WORD * version) {
    char myname[20] = "Advanced map editor"; //No more than 20 chars
    memcpy(name,myname,20);
    //WORD myversion = 0x0100; //Version(8) | Subversion(8)
    *version = 0x0100;//myversion;
    return 1;
}

// Calculates the selection box in the charset
void SetupChSelection(HWND hwnd) {
	int xFrom = chX(selACh) < chX(selBCh) ? chX(selACh) : chX(selBCh);
	int yFrom = chY(selACh) < chY(selBCh) ? chY(selACh) : chY(selBCh);
	int xTo   = chX(selACh) < chX(selBCh) ? chX(selBCh) : chX(selACh);
	int yTo   = chY(selACh) < chY(selBCh) ? chY(selBCh) : chY(selACh);
	char buf[20];

	selBeginCh  = yFrom * 32 + xFrom;
	selEndCh    = yTo * 32 + xTo;
	selWidthCh  = xTo - xFrom + 1;
	selHeightCh = yTo - yFrom + 1;

	// Update the labels on the left pane
	sprintf_s(buf, "%d", selBeginCh);
	SetWindowText(GetDlgItem(hwnd, IDC_STATICBEGCH), buf);
	sprintf_s(buf, "%d", selEndCh);
	SetWindowText(GetDlgItem(hwnd, IDC_STATICENDCH), buf);
	sprintf_s(buf, "%dx%d", selWidthCh, selHeightCh);
	SetWindowText(GetDlgItem(hwnd, IDC_STATICSELCH), buf);

	// If a single char is selected
	if (selBeginCh == selEndCh) {
		// Re-populate the animation listbox
		SendDlgItemMessage(hwnd, IDC_LISTANI, WM_SETREDRAW, FALSE, 0);
		SendDlgItemMessage(hwnd, IDC_LISTANI, LB_RESETCONTENT, 0, 0);
		aniframe *tmp = &aniframes[selBeginCh];
		while (tmp) {
			SendDlgItemMessage(hwnd, IDC_LISTANI, LB_ADDSTRING, 0, (LPARAM)tmp->charNo);
			tmp = tmp->next;
		}
		SendDlgItemMessage(hwnd, IDC_LISTANI, WM_SETREDRAW, TRUE, 0);
		EnableWindow(GetDlgItem(hwnd, IDC_LISTANI), TRUE);
	}
	// Otherwise, disable it
	else {
		SendDlgItemMessage(hwnd, IDC_LISTANI, LB_RESETCONTENT, 0, 0);
		EnableWindow(GetDlgItem(hwnd, IDC_LISTANI), FALSE);
		SendMessage(hwnd, WM_NEXTDLGCTL, (WPARAM)GetDlgItem(hwnd, IDOK), TRUE);
	}

	InvalidateRect(hwnd, &chRECT, FALSE);
}

// Calculates the selection box in the bgmap
void SetupSelection(HWND hwnd) {
	int xFrom = X(selA) < X(selB) ? X(selA) : X(selB);
	int yFrom = Y(selA) < Y(selB) ? Y(selA) : Y(selB);
	int xTo   = X(selA) < X(selB) ? X(selB) : X(selA);
	int yTo   = Y(selA) < Y(selB) ? Y(selB) : Y(selA);
	char buf[20];

	selBegin  = yFrom * 64 + xFrom;
	selEnd    = yTo * 64 + xTo;
	selWidth  = xTo - xFrom + 1;
	selHeight = yTo - yFrom + 1;

	// Update the labels on the left pane
	sprintf_s(buf, "%d", selBegin);
	SetWindowText(GetDlgItem(hwnd, IDC_STATICBEG), buf);
	sprintf_s(buf, "%d", selEnd);
	SetWindowText(GetDlgItem(hwnd, IDC_STATICEND), buf);
	sprintf_s(buf, "%dx%d", selWidth, selHeight);
	SetWindowText(GetDlgItem(hwnd, IDC_STATICSEL), buf);

	// If a single char is selected
	if (selBegin == selEnd) {
		// Show detailed info about that char
		sprintf_s(buf, "%d", bgMap[selBegin] & 0x1FF);
		SetWindowText(GetDlgItem(hwnd, IDC_STATICCHAR), buf);
		sprintf_s(buf, "%c%c", (bgMap[selBegin] & 0x2000) ? 'H' : '-', (bgMap[selBegin] & 0x1000) ? 'V' : '-');
		SetWindowText(GetDlgItem(hwnd, IDC_STATICFLIP), buf);
		sprintf_s(buf, "%d", (bgMap[selBegin] & 0xC000) >> 14);
		SetWindowText(GetDlgItem(hwnd, IDC_STATICPAL), buf);

		// If "Follow map" is checked
		if (SendMessage(GetDlgItem(hwnd, IDC_CHECKFOLLOW), BM_GETCHECK, 0, 0)) {
			// Update the selection in the charset accordingly
			selACh = selBCh = bgMap[selBegin] & 0x1FF;
			SetupChSelection(hwnd);
		}
	}
	else {
		// Show no detailed info for large selections
		SetWindowText(GetDlgItem(hwnd, IDC_STATICCHAR), "-");
		SetWindowText(GetDlgItem(hwnd, IDC_STATICFLIP), "-");
		SetWindowText(GetDlgItem(hwnd, IDC_STATICPAL), "-");
	}

	InvalidateRect(hwnd, &mapRECT, FALSE);
}

// Finds out if a specific index is inside the selection box
bool IsInSelection(int index, int start, int end, int area = 1) {
	if (index < start || index > end)
		return false;
	// The bgmap
	if (area == 1 && (X(index) < X(start) || X(index) > X(end)))
		return false;
	// The charset
	else if (area == 2 && (chX(index) < chX(start) || chX(index) > chX(end)))
		return false;
	return true;
}

// Blits a single char from a charset to the bgmap area
// Defaults to the char defined in each map cell, but this can be overridden for animations
void Char2MapBlt(int index, int aniChar = -1) {
	int charNo, pal, x, y;
	bool hFlip, vFlip;

	x = X(index);
	y = Y(index);
	charNo = aniChar > -1 ? aniChar : bgMap[index] & 0x1FF;
	hFlip  = (bgMap[index] & 0x2000) > 0;
	vFlip  = (bgMap[index] & 0x1000) > 0;
	pal    = (bgMap[index] & 0xC000) >> 14;
	StretchBlt(mapDC, hFlip ? x*8+7 : x*8, vFlip ? y*8+7 : y*8, hFlip ? -8 : 8, vFlip ? -8 : 8, charDC[pal], (charNo & 0x1F)*8, (charNo >> 5)*8, 8, 8, SRCCOPY);
}

// Saves an undo state before map modifications
void PushUndo(HWND hwnd) {
	undo *tmp = new undo;
	// Simply save a copy of the entire bgmap
	memcpy(tmp->bgMap, bgMap, 8192);
	lastUndo ? tmp->prev = lastUndo : tmp->prev = NULL;
	lastUndo = tmp;
	EnableWindow(GetDlgItem(hwnd, IDC_BTNUNDO), TRUE);
}

// Restores the previous undo state
void PopUndo(HWND hwnd) {
	if (!lastUndo)
		return;
	// Step through each map cell
	for (int i = 0; i < 4096; i++)
		// If it differs from the previous state, restore it
		if (lastUndo->bgMap[i] != bgMap[i]) {
			bgMap[i] = lastUndo->bgMap[i];
			Char2MapBlt(i);
		}
	// Delete the undo node
	undo *tmp = lastUndo;
	lastUndo = lastUndo->prev;
	delete tmp;
	SetupSelection(hwnd);
	// Gray out the button if there are no more undo nodes
	if (!lastUndo) {
		EnableWindow(GetDlgItem(hwnd, IDC_BTNUNDO), FALSE);
		SendMessage(hwnd, WM_NEXTDLGCTL, (WPARAM)GetDlgItem(hwnd, IDOK), TRUE);
	}
}

// Adds a frame to the specified char's animation sequence
void AddAniFrame(int baseChar, int nextChar) {
	aniframe *tmp = &aniframes[baseChar];
	// Find the end of the sequence
	while (tmp->next)
		tmp = tmp->next;
	// Add the new frame
	tmp->next = new aniframe;
	tmp->next->charNo = nextChar;
	tmp->next->next = NULL;
}

// Removes a frame from the specified char's animation sequence
void DeleteAniFrame(int baseChar, int idxDelete) {
	// You can never remove the first frame
	if (idxDelete < 1)
		return;
	aniframe *prior = &aniframes[baseChar];
	// Find the frame prior to the one to remove
	for (int i = 1; i < idxDelete; i++)
		prior = prior->next;
	// The next one is the one to go
	aniframe *del = prior->next;
	prior->next = del->next;
	// If the one we're deleting is the one currently displayed
	if (currFrame[baseChar] == del)
		// See to it that the next frame will be defined
		currFrame[baseChar] = prior;
	delete del;
}

// Updates a charset with new colors when you click on a palette
void UpdatePalette(int pal) {
	int i, j;
	for (i = 0; i < 128; i++)
		for (j = 0; j < 256; j++) {
			int idx = charSet[(127-i)*256+j];
			// There are four differently colored charmaps, update the right one
			SetPixel(charDC[pal], j, i, RGB(color[pal][idx], 0, 0));
		}
	// Redraw any map cells that uses this palette
	for (i = 0; i < 4096; i++)
		if (((bgMap[i] & 0xC000) >> 14) == pal)
			Char2MapBlt(i);
}

// Performs a horizontal flip on the entire selection
void HFlip(HWND hwnd) {
	// Get the center x coordinate of the selection
	unsigned int i, xCenter = (X(selBegin) + X(selEnd)) / 2;
	PushUndo(hwnd);
	for (i = selBegin; i <= selEnd; i++)
		// For all tiles contained within the selection
		if (IsInSelection(i, selBegin, selEnd))
			// If they are in the left half of the selection
			if (X(i) <= xCenter) {
				// Find their corresponding "mirror" tile in the right half
				int idxMirror = (X(selEnd) - (X(i) - X(selBegin))) - X(i);
				WORD *mirror = &bgMap[i] + idxMirror;
				WORD tmpCell = *mirror;
				// Swap positions of the left and right tiles while also flipping them
				*mirror = bgMap[i] ^ 0x2000;
				bgMap[i] = tmpCell ^ 0x2000;
				// Update the graphics on the map
				Char2MapBlt(i);
				Char2MapBlt(i + idxMirror);
			}
	SetupSelection(hwnd);
}

// Performs a vertical flip. Same principles as HFlip above
void VFlip(HWND hwnd) {
	unsigned int i, yCenter = (Y(selBegin) + Y(selEnd)) / 2;
	PushUndo(hwnd);
	for (i = selBegin; i <= selEnd; i++)
		if (IsInSelection(i, selBegin, selEnd))
			if (Y(i) <= yCenter) {
				int idxMirror = (Y(selEnd) - (Y(i) - Y(selBegin))) - Y(i);
				WORD *mirror = &bgMap[i] + idxMirror * 64;
				WORD tmpCell = *mirror;
				*mirror = bgMap[i] ^ 0x1000;
				bgMap[i] = tmpCell ^ 0x1000;
				Char2MapBlt(i);
				Char2MapBlt(i + idxMirror * 64);
			}
	SetupSelection(hwnd);
}

// Cycles through the four palettes for the entire selection
void CyclePalettes(HWND hwnd) {
	PushUndo(hwnd);
	for (unsigned int i = selBegin; i <= selEnd; i++)
		if (IsInSelection(i, selBegin, selEnd)) {
			// Read out the old palette and increase it by 1
			int pal = ((bgMap[i] & 0xC000) >> 14) + 1;
			// Clear the palette data from the cell
			bgMap[i] &= 0x3FFF;
			// Set the new palette
			bgMap[i] |= pal << 14;
			Char2MapBlt(i);
		}
	SetupSelection(hwnd);
}

// Toggles the 11th bit of each cell in the selection
void ToggleFlag(HWND hwnd) {
	PushUndo(hwnd);
	for (unsigned int i = selBegin; i <= selEnd; i++)
		if (IsInSelection(i, selBegin, selEnd))
			bgMap[i] ^= 0x800;
	InvalidateRect(hwnd, NULL, FALSE);
}

// Expands the selection to include the entire bgmap
void EditSelectAll(HWND hwnd) {
	selA = 0;
	selB = 4095;
	SetupSelection(hwnd);
}

// Sets all tiles in the selection to zero
void EditDelete(HWND hwnd) {
	PushUndo(hwnd);
	for (unsigned int i = selBegin; i <= selEnd; i++)
		if (IsInSelection(i, selBegin, selEnd)) {
			bgMap[i] = 0x0000;
			Char2MapBlt(i);
		}
	SetupSelection(hwnd);
}

// Copies the current selection (bgmap or charset) to the Windows clipboard
void EditCopy(HWND hwnd, int start, int end, int width, int height, int fromArea) {
	WORD *pGlobal;
	GLOBALHANDLE hGlobal;
	
	// Allocate enough memory (2 words per tile) for the selection data
	// + 2 words at the start to describe the width and height of the selection
	hGlobal = GlobalAlloc(GHND | GMEM_SHARE, (2 + width*height) * 2);
	pGlobal = (WORD*)GlobalLock(hGlobal);
	*pGlobal++ = width;
	*pGlobal++ = height;

	for (int i = start; i <= end; i++)
		if (IsInSelection(i, start, end, fromArea))
			// Fill up the memory with the clip data (either a mapcell or a char)
			*pGlobal++ = (fromArea == 1 ? bgMap[i] : i);

	// Put it on the clipboard in our own registered format
	GlobalUnlock(hGlobal);
	OpenClipboard(hwnd);
	EmptyClipboard();
	SetClipboardData(iFormat, hGlobal);
	CloseClipboard();
}

// A cut is the same as a copy followed by a delete
void EditCut(HWND hwnd) {
	EditCopy(hwnd, selBegin, selEnd, selWidth, selHeight, 1);
	EditDelete(hwnd);
}

// Pastes the selection on the Windows clipboard to the bgmap
void EditPaste(HWND hwnd) {
	WORD *pGlobal, *pBeginning;
	GLOBALHANDLE hGlobal;
	unsigned int i, j, clipWidth, clipHeight, index, baseIndex;

	OpenClipboard(hwnd);
	// See if there is any data on the clipboard in our special 'VIDE advanced map format'
	hGlobal = GetClipboardData(iFormat);

	// If not, don't try to paste it
	if (hGlobal == NULL) {
		CloseClipboard();
		return;
	}

	// Save the current map so we can undo later
	PushUndo(hwnd);

	pGlobal = (WORD*)GlobalLock(hGlobal);
	// Get the clip width and height (It's the first two values in the clip)
	clipWidth  = *pGlobal++;
	clipHeight = *pGlobal++;
	// Save the start of the clip data for repeating purposes
	pBeginning = pGlobal;

	baseIndex = selBegin;

	// Always paste the entire clip at least once, regardless of selection size
	do {
		// For each row...
		for (i = 0; i < clipHeight; i++) {
			// ...and each column of the clip
			for (j = 0; j < clipWidth; j++) {
				// Find the index of the current mapcell to paste over
				index = baseIndex + i*64+j;
				// Set the cell data to the cell data from the clipboard
				bgMap[index] = *pGlobal++;
				// Also update the graphics of the cell
				Char2MapBlt(index);
				// If we have reached the right edge of the map (or the selection, when repeating)...
				if ((X(index) == 63) || ((X(index) == X(selEnd)) && (baseIndex != selBegin))) {
					// Jump-skip the clipboard pointer to the end of the row
					pGlobal += clipWidth-j-1;
					// And stop pasting the rest of the current row (to avoid wrapping)
					break;
				}
			}
			// if we're at the bottom row of the map (or the selection, when repeating)...
			if ((index > 4031) || ((Y(index) == Y(selEnd)) && (baseIndex != selBegin)))
				// ...stop pasting. We don't want to write data outside the map (or the selection)
				break;
		}

		// Reset the clipboard pointer to get access to the data all over again
		pGlobal = pBeginning;
		
		// If the next copy of the clip would appear outside the selection
		if (X(baseIndex) + clipWidth > X(selEnd)) {
			// Basically do a carriage return and line feed for the clip
			baseIndex -= X(baseIndex) - X(selBegin);
			baseIndex += clipHeight * 64;
		}
		else
			// Else, make the next copy appear to the right of the current one
			baseIndex += clipWidth;
	}
	// If the clip fully fits inside the selection, repeat the pattern to fill the whole selection
	while ((selWidth >= clipWidth) && (selHeight >= clipHeight) && (baseIndex <= selEnd));

	GlobalUnlock(hGlobal);
	CloseClipboard();
	SetupSelection(hwnd);
}

// Pastes a single tile to all cells on the bgmap that
// uses the same tile as the current selection does
void EditReplaceAll(HWND hwnd) {
	WORD *pGlobal;
	GLOBALHANDLE hGlobal;

	OpenClipboard(hwnd);
	// See if there is any data on the clipboard in our special 'VIDE advanced map format'
	hGlobal = GetClipboardData(iFormat);

	// If not, don't try to paste it
	if (hGlobal == NULL) {
		CloseClipboard();
		return;
	}

	pGlobal = (WORD*)GlobalLock(hGlobal);
	// Get the first actual char in the clip
	int clipChar = *(pGlobal + 2) & 0x1FF;
	GlobalUnlock(hGlobal);
	CloseClipboard();

	// Save the current map so we can undo later
	PushUndo(hwnd);

	int selChar = bgMap[selBegin] & 0x1FF;

	// For every cell in the bgmap
	for (int i = 0; i < 4096; i++)
		// If it uses the selected char
		if ((bgMap[i] & 0x1FF) == selChar) {
			// Set it to the clipboard's char instead, while keeping its other properties
			bgMap[i] = (bgMap[i] & ~0x1FF) | clipChar;
			// Also update the graphics of the cell
			Char2MapBlt(i);
		}
	SetupSelection(hwnd);
}

// The windows procedure for the dialog
BOOL CALLBACK AdvMapDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam) {
	HDC            hdc;
	PAINTSTRUCT    ps;
	RECT           rect;
	static HPEN    hPurplePen, hDPurplPen;
	static HBRUSH  hPalBrush[4];
	static HFONT   hFont;
	static HDC     dblBufDC, dblBufChDC;
	static HBITMAP dblBufBMP, dblBufChBMP;
	static int     currArea = 1;
	int			   i, j;
	bool           shiftUp;

	switch(Message)
	{
		case WM_INITDIALOG:
			// Set up icon and text in the title bar
			SendMessage(hwnd, WM_SETICON, ICON_SMALL, (LPARAM)LoadImage(hDllModule, MAKEINTRESOURCE(IDI_CHARMAP), IMAGE_ICON, 16, 16, 0));
			sprintf_s(mapName, "%s - Advanced map editor", mapName);
			SetWindowText(hwnd, mapName);
			// Register our own clipboard format
			iFormat = RegisterClipboardFormat("CF_VIDEADVMAP");
			// Set up the icons for all the toolbar buttons
			SendDlgItemMessage(hwnd, IDC_BTNUNDO, BM_SETIMAGE, IMAGE_ICON, (LPARAM)LoadImage(hDllModule, MAKEINTRESOURCE(IDI_UNDO), IMAGE_ICON, 16, 16, 0));
			SendDlgItemMessage(hwnd, IDC_BTNCUT, BM_SETIMAGE, IMAGE_ICON, (LPARAM)LoadImage(hDllModule, MAKEINTRESOURCE(IDI_CUT), IMAGE_ICON, 16, 16, 0));
			SendDlgItemMessage(hwnd, IDC_BTNCOPY, BM_SETIMAGE, IMAGE_ICON, (LPARAM)LoadImage(hDllModule, MAKEINTRESOURCE(IDI_COPY), IMAGE_ICON, 16, 16, 0));
			SendDlgItemMessage(hwnd, IDC_BTNPASTE, BM_SETIMAGE, IMAGE_ICON, (LPARAM)LoadImage(hDllModule, MAKEINTRESOURCE(IDI_PASTE), IMAGE_ICON, 16, 16, 0));
			SendDlgItemMessage(hwnd, IDC_BTNREPLACE, BM_SETIMAGE, IMAGE_ICON, (LPARAM)LoadImage(hDllModule, MAKEINTRESOURCE(IDI_REPLACEALL), IMAGE_ICON, 16, 16, 0));
			SendDlgItemMessage(hwnd, IDC_BTNDELETE, BM_SETIMAGE, IMAGE_ICON, (LPARAM)LoadImage(hDllModule, MAKEINTRESOURCE(IDI_DELETE), IMAGE_ICON, 16, 16, 0));
			SendDlgItemMessage(hwnd, IDC_BTNVFLIP, BM_SETIMAGE, IMAGE_ICON, (LPARAM)LoadImage(hDllModule, MAKEINTRESOURCE(IDI_VFLIP), IMAGE_ICON, 16, 16, 0));
			SendDlgItemMessage(hwnd, IDC_BTNHFLIP, BM_SETIMAGE, IMAGE_ICON, (LPARAM)LoadImage(hDllModule, MAKEINTRESOURCE(IDI_HFLIP), IMAGE_ICON, 16, 16, 0));
			SendDlgItemMessage(hwnd, IDC_BTNPAL, BM_SETIMAGE, IMAGE_ICON, (LPARAM)LoadImage(hDllModule, MAKEINTRESOURCE(IDI_PALETTE), IMAGE_ICON, 16, 16, 0));
			SendDlgItemMessage(hwnd, IDC_BTNFLAG, BM_SETIMAGE, IMAGE_ICON, (LPARAM)LoadImage(hDllModule, MAKEINTRESOURCE(IDI_FLAG), IMAGE_ICON, 16, 16, 0));
			SendDlgItemMessage(hwnd, IDC_CHECKFOLLOW, BM_SETCHECK, 1, 0) ;
			SendDlgItemMessage(hwnd, IDC_SPINMS, UDM_SETRANGE, 0, (LPARAM)1000);
			SendDlgItemMessage(hwnd, IDC_SPINMS, UDM_SETPOS, 0, (LPARAM)200);
			SendDlgItemMessage(hwnd, IDC_LISTANI, LB_SETITEMHEIGHT, 0, (LPARAM)20);

			hdc = GetDC(hwnd); 
			// Create five device contexts for the charsets (four palettes + black 'n white)
			for (i = 0; i < 5; i++) {
				charDC[i]  = CreateCompatibleDC(hdc);
				charBMP[i] = CreateCompatibleBitmap(hdc, 256, 128);
				SelectObject(charDC[i], charBMP[i]);
			}

			// Create the area to hold the bgmap graphics
			mapDC  = CreateCompatibleDC(hdc);
			mapBMP = CreateCompatibleBitmap(hdc, 512, 512);
			SelectObject(mapDC, mapBMP);

			// A double-buffer for the charset area
			dblBufChDC  = CreateCompatibleDC(hdc);
			dblBufChBMP = CreateCompatibleBitmap(hdc, 512, 256);
			SelectObject(dblBufChDC, dblBufChBMP);

			// A double-buffer for the bgmap area
			dblBufDC  = CreateCompatibleDC(hdc);
			dblBufBMP = CreateCompatibleBitmap(hdc, 512, 512);
			SelectObject(dblBufDC, dblBufBMP);
			hFont = CreateFont(10, 0, 0, 0, 0, 0, 0, 0, 0, OUT_RASTER_PRECIS, 0, 0, 0, TEXT("Small Fonts"));
			SelectObject(dblBufDC, hFont);
			SetBkMode(dblBufDC, TRANSPARENT);
			SetTextAlign(dblBufDC, TA_CENTER);
			SetTextColor(dblBufDC, RGB(0, 0, 255));

			// Some nice pens to draw the selection boxes with
			hPurplePen = CreatePen(PS_SOLID, 1, RGB(155, 0, 255));
			hDPurplPen = CreatePen(PS_SOLID, 1, RGB(100, 0, 150));
			// Four red brushes to paint the palettes with
			for (i = 0; i < 4; i++)
				hPalBrush[i] = CreateSolidBrush(RGB(i*85, 0, 0));			

			// Draw the five charset areas in different colors
			// These will always reside off-screen
			// (remember, one for each palette + one b/w)
			for (i = 0; i < 128; i++)
				for (j = 0; j < 256; j++) {
					int idx = charSet[(127-i)*256+j];
					SetPixel(charDC[0], j, i, RGB(color[0][idx], 0, 0));
					SetPixel(charDC[1], j, i, RGB(color[1][idx], 0, 0));
					SetPixel(charDC[2], j, i, RGB(color[2][idx], 0, 0));
					SetPixel(charDC[3], j, i, RGB(color[3][idx], 0, 0));
					SetPixel(charDC[4], j, i, RGB(idx*85, idx*85, idx*85));
				}

			// Set up the bgmap using the charsets we just drew
			for (i = 0; i < 4096; i++)
				Char2MapBlt(i);
			
			ReleaseDC(hwnd, hdc);

			// Initialize the animation setup (no animations to begin with)
			for (i = 0; i < 512; i++) {
				aniframes[i].charNo = i;
				aniframes[i].next = NULL;
				currFrame[i] = &aniframes[i];
			}

			// Put the selection boxes in the upper left corners
			selA = selB = 0;
			selACh = selBCh = 0;
			lastUndo = NULL;
			SetupSelection(hwnd);
			SetupChSelection(hwnd);

		return TRUE;

		case WM_LBUTTONDBLCLK:
		case WM_LBUTTONDOWN:
			// On the map
			if (LOWORD(lParam) > 63 && HIWORD(lParam) < 512) {
				// Calculate the index of the cell that was clicked
				selB = (HIWORD(lParam) & ~7) * 8 + ((LOWORD(lParam) & ~7) - 64) / 8;
				// If shift is pressed, expand the selection. Otherwise just move it
				if (!(wParam & MK_SHIFT))
					selA = selB;
				currArea = 1;
				SetupSelection(hwnd);
			}
			// On the chars
			else if (LOWORD(lParam) > 63 && HIWORD(lParam) > 513) {
				selBCh = ((HIWORD(lParam) - 514) & ~15) * 2 + ((LOWORD(lParam) & ~15) - 64) / 16;
				if (!(wParam & MK_SHIFT))
					selACh = selBCh;
				currArea = 2;
				SetupChSelection(hwnd);
			}
			// On the palettes
			else if (LOWORD(lParam) > 13 && LOWORD(lParam) < 57 && HIWORD(lParam) > 283 && HIWORD(lParam) < 327) {
				// Cycle through the four shades on any square you click
				i = (HIWORD(lParam) - 284) / 11;
				j = (LOWORD(lParam) - 14) / 11;
				if ((color[i][3-j] += 85) < 85)
					color[i][3-j] = 0;
				UpdatePalette(i);
				InvalidateRect(hwnd, NULL, FALSE);
			}
		return TRUE;

		case WM_RBUTTONDOWN:
			// Right-click on a char adds it to the current char's animation
			if (LOWORD(lParam) > 63 && HIWORD(lParam) > 513 && selBeginCh == selEndCh) {
				int newFrame = ((HIWORD(lParam) - 514) & ~15) * 2 + ((LOWORD(lParam) & ~15) - 64) / 16;
				AddAniFrame(selBeginCh, newFrame);
				SetupChSelection(hwnd);
			}
		return TRUE;

		case WM_MOUSEMOVE:
			if (wParam & MK_LBUTTON && LOWORD(lParam) > 63) {
				// On the map
				if (HIWORD(lParam) < 512 && currArea == 1) {
					// Resize the selection box
					selB = (HIWORD(lParam) & ~7) * 8 + ((LOWORD(lParam) & ~7) - 64) / 8;
					SetupSelection(hwnd);
				}
				// On the chars
				else if (HIWORD(lParam) > 513 && currArea == 2) {
					selBCh = ((HIWORD(lParam) - 514) & ~15) * 2 + ((LOWORD(lParam) & ~15) - 64) / 16;
					SetupChSelection(hwnd);
				}
			}
		return TRUE;

		case WM_PAINT:
			// Blit the map to its double-buffer
			BitBlt(dblBufDC, 0, 0, 512, 512, mapDC, 0, 0, SRCCOPY);

			// Blit the chars to their double-buffer
			StretchBlt(dblBufChDC, 0, 0, 512, 256, charDC[4], 0, 0, 256, 128, SRCCOPY);

			// If "Show Flip" is checked
			if (IsDlgButtonChecked(hwnd, IDC_CHECKFLIP)) {
				SelectObject(dblBufDC, GetStockObject(WHITE_PEN));
				for (i = 0; i < 64; i++)
					for (j = 0; j < 64; j++) {
						// If a cell is H-Flipped, draw an arrow on it
						if (bgMap[i*64+j] & 0x2000) {
							MoveToEx(dblBufDC, j*8, i*8 + 3, NULL);
							LineTo  (dblBufDC, j*8 + 7, i*8 + 3);
							MoveToEx(dblBufDC, j*8 + 1, i*8 + 2, NULL);
							LineTo  (dblBufDC, j*8 + 1, i*8 + 5);
							MoveToEx(dblBufDC, j*8 + 5, i*8 + 2, NULL);
							LineTo  (dblBufDC, j*8 + 5, i*8 + 5);
						}
						// If a cell is V-Flipped, draw an arrow on it
						if (bgMap[i*64+j] & 0x1000) {
							MoveToEx(dblBufDC, j*8 + 3, i*8, NULL);
							LineTo  (dblBufDC, j*8 + 3, i*8 + 7);
							MoveToEx(dblBufDC, j*8 + 2, i*8 + 1, NULL);
							LineTo  (dblBufDC, j*8 + 5, i*8 + 1);
							MoveToEx(dblBufDC, j*8 + 2, i*8 + 5, NULL);
							LineTo  (dblBufDC, j*8 + 5, i*8 + 5);
						}
					}
			}

			// If "Show Flag" is checked
			if (IsDlgButtonChecked(hwnd, IDC_CHECKFLAG)) {
				for (i = 0; i < 64; i++)
					for (j = 0; j < 64; j++)
						// If a cell's bit 11 is set, make it blueish
						if (bgMap[i*64+j] & 0x800)
							BitBlt(dblBufDC, j*8, i*8, 8, 8, mapDC, j*8, i*8, DSTINVERT);
			}

			// If "Show Palette" is checked
			if (IsDlgButtonChecked(hwnd, IDC_CHECKPAL)) {
				int pal;
				char buf[2];
				for (i = 0; i < 64; i++)
					for (j = 0; j < 64; j++) {
						// If greater than 0, display the palette number on each map cell
						if (pal = (bgMap[i*64+j] & 0xC000) >> 14) {
							sprintf_s(buf, "%d", pal);
							TextOut(dblBufDC, 4 + j*8, i*8 - 1, buf, 1);
						}
					}
			}

			// Draw the map selection box
			MoveToEx(dblBufDC, mapX(selBegin), mapY(selEnd)+6, NULL);
			SelectObject(dblBufDC, hPurplePen);
			LineTo(dblBufDC, mapX(selBegin), mapY(selBegin));
			LineTo(dblBufDC, mapX(selEnd)+7, mapY(selBegin));
			SelectObject(dblBufDC, hDPurplPen);
			LineTo(dblBufDC, mapX(selEnd)+7, mapY(selEnd)+7);
			LineTo(dblBufDC, mapX(selBegin)-1, mapY(selEnd)+7);

			// Draw the char selection box
			MoveToEx(dblBufChDC, chmapX(selBeginCh), chmapY(selEndCh)+14, NULL);
			SelectObject(dblBufChDC, hPurplePen);
			LineTo(dblBufChDC, chmapX(selBeginCh), chmapY(selBeginCh));
			LineTo(dblBufChDC, chmapX(selEndCh)+15, chmapY(selBeginCh));
			SelectObject(dblBufChDC, hDPurplPen);
			LineTo(dblBufChDC, chmapX(selEndCh)+15, chmapY(selEndCh)+15);
			LineTo(dblBufChDC, chmapX(selBeginCh)-1, chmapY(selEndCh)+15);

			hdc = BeginPaint(hwnd, &ps);
			GetClientRect(hwnd, &rect);

			// Blit the double-buffers to the screen
			BitBlt(hdc, 64, 0, 512, 512, dblBufDC, 0, 0, SRCCOPY);
			BitBlt(hdc, 64, 514, 512, 256, dblBufChDC, 0, 0, SRCCOPY);

			// Draw the palettes
			SelectObject(hdc, GetStockObject(NULL_PEN));
			for (i = 0; i < 4; i++)
				for (j = 0; j < 4; j++) {
					SelectObject(hdc, hPalBrush[color[i][3-j] / 85]);
					Rectangle(hdc, 14 + j*11, 284 + i*11, 25 + j*11, 295 + i*11);
				}
			EndPaint(hwnd, &ps);
		return 0;

		case WM_KEYDOWN:
			// Handles the arrow keys to move/resize the selection boxes
			shiftUp = GetKeyState(VK_SHIFT) >= 0;
			switch (wParam)
			{
				case VK_LEFT:
					// On the bgmap
					if (currArea == 1 && (shiftUp || X(selB) > 0))
						selB = --selB & 0xFFF;
					// On the charset
					else if (currArea == 2 && (shiftUp || chX(selBCh) > 0))
						selBCh = --selBCh & 0x1FF;
				break;
				case VK_RIGHT:
					if (currArea == 1 && (shiftUp || X(selB) < 63))
						selB = ++selB & 0xFFF;
					else if (currArea == 2 && (shiftUp || chX(selBCh) < 31))
						selBCh = ++selBCh & 0x1FF;
				break;
				case VK_UP:
					if (currArea == 1 && (shiftUp || Y(selB) > 0))
						selB = (selB-64) & 0xFFF;
					else if (currArea == 2 && (shiftUp || chY(selBCh) > 0))
						selBCh = (selBCh-32) & 0x1FF;
				break;
				case VK_DOWN:
					if (currArea == 1 && (shiftUp || Y(selB) < 63))
						selB = (selB+64) & 0xFFF;
					else if (currArea == 2 && (shiftUp || chY(selBCh) < 15))
						selBCh = (selBCh+32) & 0x1FF;
				break;
				case VK_TAB:
					// Tab shifts focus between the bgmap and the charset areas
					currArea == 1 ? currArea = 2 : currArea = 1;
				break;
				case VK_DELETE:
					EditDelete(hwnd);
				break;
			}
			// If it's an arrow key
			if (wParam > 0x24 && wParam < 0x29) {
				if (currArea == 1){
					if (shiftUp)
						// Move the selection instead of just resizing it
						selA = selB;
					SetupSelection(hwnd);
				}
				else if (currArea == 2){
					if (shiftUp)
						selACh = selBCh;
					SetupChSelection(hwnd);
				}
			}
		return 0;

		case WM_CHAR:
			// The keyboard hotkeys
			switch (wParam)
			{
				case 0x01: // Ctrl+A
					EditSelectAll(hwnd);
				break;
				case 0x18: // Ctrl+X
					EditCut(hwnd);
				break;
				case 0x03: // Ctrl+C
					if (currArea == 1) // Copy from the bgmap
						EditCopy(hwnd, selBegin, selEnd, selWidth, selHeight, currArea);
					else if (currArea == 2) // Copy from the charset
						EditCopy(hwnd, selBeginCh, selEndCh, selWidthCh, selHeightCh, currArea);
				break;
				case 0x16: // Ctrl+V
					EditPaste(hwnd);
				break;
				case 0x12: // Ctrl+R
					EditReplaceAll(hwnd);
				break;
				case 0x08: // Ctrl+H
					HFlip(hwnd);
				break;
				case 0x15: // Ctrl+U
					VFlip(hwnd);
				break;
				case 0x10: // Ctrl+P
					CyclePalettes(hwnd);
				break;
				case 0x06: // Ctrl+F
					ToggleFlag(hwnd);
				break;
				case 0x1A: // Ctrl+Z
					PopUndo(hwnd);
				break;
			}
		return 0;

		case WM_COMMAND:
			// The buttons and other window controls
			switch(LOWORD(wParam))
			{
				case IDC_BTNUNDO:
					PopUndo(hwnd);
				break;
				case IDC_BTNCUT:
					EditCut(hwnd);
				break;
				case IDC_BTNCOPY:
					if (currArea == 1) // Copy from the bgmap
						EditCopy(hwnd, selBegin, selEnd, selWidth, selHeight, currArea);
					else if (currArea == 2) // Copy from the charset
						EditCopy(hwnd, selBeginCh, selEndCh, selWidthCh, selHeightCh, currArea);
				break;
				case IDC_BTNPASTE:
					EditPaste(hwnd);
				break;
				case IDC_BTNREPLACE:
					EditReplaceAll(hwnd);
				break;
				case IDC_BTNDELETE:
					EditDelete(hwnd);
				break;
				case IDC_BTNVFLIP:
					VFlip(hwnd);
				break;
				case IDC_BTNHFLIP:
					HFlip(hwnd);
				break;
				case IDC_BTNPAL:
					CyclePalettes(hwnd);
				break;
				case IDC_BTNFLAG:
					ToggleFlag(hwnd);
				break;
				case IDC_CHECKFLIP:
				case IDC_CHECKPAL:
				case IDC_CHECKFLAG:
					// The "Show" checkboxes just redraws the screen
					InvalidateRect(hwnd, NULL, FALSE);
				break;
				case IDC_CHECKANI:
					if (SendMessage((HWND)lParam, BM_GETCHECK, 0, 0))
						// Start the animation at the specified timer interval
						SetTimer(hwnd, TIMER_ANI, SendDlgItemMessage(hwnd, IDC_SPINMS, UDM_GETPOS, 0, 0), NULL);
					else {
						// Stop the animation and reset the bgmap to its first frame
						KillTimer(hwnd, TIMER_ANI);
						// For every char in the charset
						for (i = 0; i < 512; i++)
							// If it has an animation setup
							if (aniframes[i].next) {
								// Reset the animation
								currFrame[i] = &aniframes[i];
								// For every cell in the bgmap
								for (j = 0; j < 4096; j++)
									// If it uses the current char
									if ((bgMap[j] & 0x1FF) == i)
										// Reset its original char
										Char2MapBlt(j);
							}
						InvalidateRect(hwnd, &mapRECT, FALSE);
					}
				break;
				case IDC_LISTANI:
					// Doubleclicking in the listbox removes animation frames
					if (HIWORD(wParam) == LBN_DBLCLK) {
						int idx = SendMessage((HWND)lParam, LB_GETCURSEL, 0, 0);
						DeleteAniFrame(selBeginCh, idx);
						// If the char no longer has an animation
						if (!aniframes[selBeginCh].next)
							// For every cell in the bgmap
								for (j = 0; j < 4096; j++)
									// If it uses the current char
									if ((bgMap[j] & 0x1FF) == selBeginCh)
										// Reset its original char
										Char2MapBlt(j);
						SetupChSelection(hwnd);
					}
				break;
				case IDOK:
					dlgResult = IDOK;
					DestroyWindow(hwnd);
					//EndDialog(hwnd, IDOK);
				break;
				case IDCANCEL:
					dlgResult = IDCANCEL;
					DestroyWindow(hwnd);
					//EndDialog(hwnd, IDCANCEL);
				break;
			}
		break;

		case WM_TIMER:
			// The timer is what drives the animation.
			// For every char in the charset
			for (i = 0; i < 512; i++) {
				// If it has an animation setup
				if (aniframes[i].next) {
					// Advance to its next frame
					currFrame[i] = currFrame[i]->next ? currFrame[i]->next : &aniframes[i];
					
					// For every cell in the bgmap
					for (j = 0; j < 4096; j++)
						// If it uses the current char
						if ((bgMap[j] & 0x1FF) == i)
							// Display its next frame instead
							Char2MapBlt(j, currFrame[i]->charNo);
				}
			}
			InvalidateRect(hwnd, &mapRECT, FALSE);
		return 0;

		case WM_VSCROLL:
			// Adjust the timer interval in ms by using the spinbutton
			if (SendDlgItemMessage(hwnd, IDC_CHECKANI, BM_GETCHECK, 0, 0))
				SetTimer(hwnd, TIMER_ANI, SendDlgItemMessage(hwnd, IDC_SPINMS, UDM_GETPOS, 0, 0), NULL);
		return 0;

		case WM_DRAWITEM: {
			// The animation listbox is owner-drawn to allow displaying the tiles in it
			DRAWITEMSTRUCT *pdis = (DRAWITEMSTRUCT*)lParam;
			HBRUSH hbrush;
			RECT rc = pdis->rcItem;
			// NULL object
			if (pdis->itemID == -1)
				return 0; 

			// Draw different stuff depending on the item's state
			switch (pdis->itemAction)
			{
				case ODA_DRAWENTIRE:
				case ODA_SELECT:
					if (pdis->itemState & ODS_SELECTED) {
						// Selected
						hbrush = CreateSolidBrush(GetSysColor(COLOR_HIGHLIGHT));
						FillRect(pdis->hDC, &rc, hbrush);
						DeleteObject(hbrush);
					}
					else {
						// Not selected
						hbrush = CreateSolidBrush(GetSysColor(COLOR_WINDOW));
						FillRect(pdis->hDC, &rc, hbrush);
						DeleteObject(hbrush);
					}
					InflateRect(&rc, -2, -2);
					StretchBlt(pdis->hDC, rc.left, rc.top, 16, 16, charDC[4], (pdis->itemData & 0x1F)*8, (pdis->itemData >> 5)*8, 8, 8, SRCCOPY);
				break;
				case ODA_FOCUS:
					// Has focus
					DrawFocusRect(pdis->hDC, &rc);
				break;
				default:
				break;
			}
		return TRUE;
		}

		case WM_DESTROY:
			// Clean up all created GDI objects before the window closes
			KillTimer(hwnd, TIMER_ANI);
			for (i = 0; i < 5; i++) {
				DeleteDC(charDC[i]);
				DeleteObject(charBMP[i]);
			}
			DeleteDC(mapDC);
			DeleteObject(mapBMP);
			DeleteDC(dblBufChDC);
			DeleteObject(dblBufChBMP);
			DeleteDC(dblBufDC);
			DeleteObject(dblBufBMP);
			DeleteObject(hPurplePen);
			DeleteObject(hDPurplPen);
			for (i = 0; i < 4; i++)
				DeleteObject(hPalBrush[i]);
			DeleteObject(hFont);
			PostQuitMessage(0);
		return 0;
		default:
			return FALSE;
	}
	return TRUE;
}

///////////////////////////////////////////////////////////////////
// Plugin Start Function
// pSet: Pointer to a 256x128 image - Type: BYTE Image[128][256] - NULL: no Charset active
// pMap: Pointer to a 4096 bgmap    - Type; WORD BGMap[4096]     - NULL: no Charmap active
// hValidWnd: Handle of the owner window to perform window updates.
///////////////////////////////////////////////////////////////////
// The Charset image only support 4 values, 0x00, 0x55, 0xAA, 0xFF
///////////////////////////////////////////////////////////////////
// BGMAP WORD Map structure
// | 15 | 14 | 13    | 12    | 11 | 10 | 09 | 08 | 07 | 06 | 05 | 04 | 03 | 02 | 01 | 00 |
// | GPLTX   | HFlip | VFlip | CHARACTER NUMBER IN SET - From 0, to 0x1FF                |

int PluginStart(BYTE * pSet, WORD * pMap, HWND hValidWnd) {
	int i;
	BYTE *pGlobal;
	GLOBALHANDLE hGlobal;

	if (!hValidWnd)
		return 0;
	if (!pSet && !pMap)
		MessageBox(NULL, "Plugin only works with Charsets or Charmaps", "Advanced map editor", MB_OK|MB_ICONASTERISK);

	// If the plugin is run on a charset, copy it to the Windows clipboard
	if (pSet) {
		// Allocate enough memory for the charset + a null terminating char
		hGlobal = GlobalAlloc(GHND | GMEM_SHARE, 256*128 + 1);
		pGlobal = (BYTE*)GlobalLock(hGlobal);

		// Fill up the memory with the data VIDE gives us, pixel by pixel
		for (i = 0; i < 256*128; i++)
			// Since we're putting this on the clipboard in plain text format,
			// add 1 to each value to avoid null characters in the middle of the string.
			// Thus, possible values for each pixel will be 1, 2, 3 or 4
			*pGlobal++ = *pSet++ / 85 + 1;

		GlobalUnlock(hGlobal);
		OpenClipboard(hValidWnd);
		EmptyClipboard();
		// Put the charset on the clipboard as a text string
		SetClipboardData(CF_TEXT, hGlobal);
		CloseClipboard();

		MessageBox(NULL, "Charset copied to clipboard. You can now run the plugin on a Charmap.", "Advanced map editor", MB_OK|MB_ICONINFORMATION);
	}

	// If the plugin is run on a charmap
	else if (pMap) {
		OpenClipboard(hValidWnd);
		hGlobal = GetClipboardData(CF_TEXT);
		pGlobal = (BYTE*)GlobalLock(hGlobal);

		// Check to see if there is a text string on the clipboard of the exact
		// length of a charset. If so, we just assume it is a charset.
		if (hGlobal == NULL || strlen((const char*)pGlobal) != 32768) {
			GlobalUnlock(hGlobal);
			CloseClipboard();
			MessageBox(NULL, "There doesn't seem to be a valid Charset in the clipboard. Please run this plugin on the associated Charset first.", "Advanced map editor", MB_OK|MB_ICONERROR);
			return 1;
		}
	  
		pSet = charSet;
		// Read back the charset from the clipboard,
		// deducting that extra 1 we added in the process
		while (*pSet = *pGlobal++)
			(*pSet++)--;
		GlobalUnlock(hGlobal);
		CloseClipboard();

		// Then copy the charmap that VIDE gives us directly
		for (i = 0; i < 4096; i++)
			bgMap[i] = *(pMap + i);

		// Get the map name, for the window titlebar
		GetWindowText(hValidWnd, mapName, 32);

		dlgResult = IDCANCEL;
		// Create the editor window and show it
		HWND hDlg = CreateDialog(hDllModule, MAKEINTRESOURCE(IDD_DIALOGMAP), hValidWnd, AdvMapDlgProc);
		ShowWindow(hDlg, SW_SHOW);

		MSG msg;
		// Send keyboard messages to the dialog window
		while (GetMessage(&msg, NULL, 0, 0)) {
			if (msg.message == WM_KEYDOWN || msg.message == WM_CHAR)
				SendMessage(hDlg, msg.message, msg.wParam, msg.lParam);

			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}

		// If the OK button was pressed to close the dialog
		if (dlgResult == IDOK) {
			for (i = 0; i < 4096; i++)
				// Transfer the updated charmap back to VIDE
				*(pMap + i) = bgMap[i];

			InvalidateRect(hValidWnd, NULL, FALSE);
		}

		// Clean up any undo points
		undo *tmp;
		while (lastUndo) {
			tmp = lastUndo;
			lastUndo = lastUndo->prev;
			delete tmp;
		}

		// Clean up any animation frames
		for (i = 0; i < 512; i++)
			while (aniframes[i].next)
				DeleteAniFrame(i, 1);
	  
	  	  
	}
	// Return to VIDE
	return 1;
}
